#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>

#include "m.h"
#include "g.h"
#include "fit.h"

static char rcsid[] = {"$Id: fit1dlsq.c,v 1.2 1997/12/15 18:52:46 kraemer Exp $"};

/*
   --- Performs least squares fit

       For algorithm see:
       Bevington, data reduction and error analysis, p. 237-239

       /mk/ 10/02/87

   Copyright  (C) 1987-1997 M.Kraemer, D-64283 Darmstadt, Hoelgesstr. 16,
                             all rights reserved
   $Log: fit1dlsq.c,v $
   Revision 1.2  1997/12/15 18:52:46  kraemer
   fit statistics in function structure

   Revision 1.1.1.1  1997/07/10 17:59:57  kraemer
   revision for cvs

*/

int fit1dlsq( struct SFITDATA    *psFD,
              struct SFITFCTBLK *psFFB,
              struct SFITSTAT   *psFS,
              unsigned long lFlag
            ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__, *pMsg = pDbgMsg;

   unsigned long lFlagCurve = 0;

   struct SFITPARBLK sFPB;
   struct SFITFCT *psFF;
   struct SFITPAR sFPRef;
   int ip1, ip2, jf, jp1, jp2;
   double *pdA, *pdC, *pdD;
   double dd, dChisqOld;
   int iLoop;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
   if ( psFS->lStatus & FITSTAT_POISSON ) lFlagCurve |= FITFCTBLK_POISSON;
   memset( &sFPB, 0, sizeof(sFPB) );
                                      /* check valid inputs */
   if ( psFD  == NULL ) goto gError;
   if ( psFFB == NULL ) goto gError;
                                      /* save parameters */
   if ( ( iRC = fitFctBlkParBlkGet(psFFB,&sFPB,lFlag) ) != 0 ) goto gError;
                                      /* compute function & derivs */
   if ( ( iRC = fitDataFctBlkDeriv(psFD,psFFB,lFlag) ) != 0 ) goto gError;
                                      /* for fixed parameters: deriv=0 */
   psFF = psFFB->psFF;
   for ( jf=0; jf<psFFB->nf; jf++ ) {
      pdD = psFF->sFPD.pd;
      for ( ip1=0; ip1<psFF->np; ip1++ ) {
         if ( psFF->plFlag[ip1] & FITPAR_FIX ) {
            vfill( 0E0, pdD, 1, psFF->sFPD.nd );
         }
         pdD += psFF->sFPD.nd;
      }
      psFF++;
   }
                                      /* compute chi-square / likelihood */
   if ( ( iRC = fitDataStatChi    (psFD,psFS,lFlag) ) != 0 ) goto gError;
                                      /* compute curvature matrices */
   if ( ( iRC = fitDataFctBlkCurv ( psFD, psFFB, lFlag | lFlagCurve ) ) != 0 ) goto gError;
                                      /* scaling w/ diagonal elements */
   pdA = psFFB->pdAlpha;
   for ( ip1=0; ip1<psFFB->np; ip1++ ) {
      if ( pdA[ip1] != 0E0 ) psFFB->pdDiag[ip1] = 1E0 / sqrt( pdA[ip1] );
      pdA += psFFB->np;
   }
   pdA = psFFB->pdAlpha;
   for ( ip1=0; ip1<psFFB->np; ip1++ ) {
      for ( ip2=0; ip2<psFFB->np; ip2++ ) {
         *pdA *= psFFB->pdDiag[ip1] * psFFB->pdDiag[ip2]; pdA++;
      }
   }
                                      /* loop limit to avoid lambda oflo */
   iLoop = ( (int) log10( psFS->dLambda ) ) + 1;
   iLoop = DBL_MAX_10_EXP - iLoop;
   do {
                                      /* scale curvature with lambda */
      vmov( psFFB->pdAlpha, 1, psFFB->pdCovar, 1, psFFB->np * psFFB->np );
      pdC = psFFB->pdCovar;
      for ( ip1=0; ip1<psFFB->np; ip1++ ) {
         *pdC = 1E0 + psFS->dLambda; pdC += psFFB->np+1;
      }
                                      /* invert curvature matrix */
      if ( InvMat( psFFB->pdCovar,
                   psFFB->np, (int *) psFFB->plRow, (int *) psFFB->plCol,
                   psFFB->pdCovar ) != 0 ) {
         iRC = ERRFIT_INVERTCURVATURE; goto gError;
      }

                                      /* update parameters, varying first */
      pdC = psFFB->pdCovar; psFF = psFFB->psFF; jp1=0;
      for ( jf=0; jf<psFFB->nf; jf++ ) {
         for ( ip1=0; ip1<psFF->np; ip1++ ) {
                                      /* accumulate parameter increment */
            if ( ! ( psFF->plFlag[ip1] & FITPAR_FIX ) ) {
               dd = 0E0;
               for ( jp2=0; jp2<psFFB->np; jp2++ ) {
                  dd += pdC[jp2] * psFFB->pdDiag[jp2] * psFFB->pdBeta[jp2];
               }
               dd *= psFFB->pdDiag[jp1];
               psFF->pdVal[ip1] = max( psFF->pdVal[ip1] + dd, psFF->pdLim[ip1*2+0] );
               psFF->pdVal[ip1] = min( psFF->pdVal[ip1]     , psFF->pdLim[ip1*2+1] );
            }
            pdC += psFFB->np; jp1++;
         }
         psFF++;
      }
                                      /* update parameters, relatively fixed */
      psFF = psFFB->psFF; jp1=0;
      for ( jf=0; jf<psFFB->nf; jf++ ) {
         for ( ip1=0; ip1<psFF->np; ip1++ ) {
            if ( ( psFF->plFlag[ip1] & FITPAR_FIX ) &&
                 ( psFFB->plRefTyp[jp1] > 0 )
               ) {
                                      /* fetch reference parameter */
               if ( fitFctBlkParGetByNum( psFFB, psFFB->plRefNum[jp1]+1, &sFPRef, lFlag|FIT_NOMSG ) ) goto gError;

               switch(psFFB->plRefTyp[jp1]) {
               case FITPAR_SUM:
                  psFF->pdVal[ip1] = psFF->pdRefVal[ip1] - sFPRef.dVal;
               break;
               case FITPAR_DIFF:
                  psFF->pdVal[ip1] = psFF->pdRefVal[ip1] + sFPRef.dVal;
               break;
               case FITPAR_PROD:
                  psFF->pdVal[ip1] = psFF->pdRefVal[ip1] / sFPRef.dVal;
               break;
               case FITPAR_QUOT:
                  psFF->pdVal[ip1] = psFF->pdRefVal[ip1] * sFPRef.dVal;
               break;
               default:
               break;
               }

               if ( psFF->pdVal[ip1] < psFF->pdLim[ip1*2+0] ) {
                  iRC = ERRFIT_DIVERGENCE; goto gError;
               }
               if ( psFF->pdVal[ip1] > psFF->pdLim[ip1*2+1] ) {
                  iRC = ERRFIT_DIVERGENCE; goto gError;
               }
            }
            jp1++;
         }
         psFF++;
      }
                                      /* compute new function values */
      if ( ( iRC = fitDataFctBlkValue(psFD,psFFB,lFlag) ) != 0 ) goto gError;
                                      /* new chi-square */
      dChisqOld = psFS->dChisq;
      if ( ( iRC = fitDataStatChi    (psFD,psFS,lFlag) ) != 0 ) goto gError;
                                      /* chisquare decreases ? */
      if ( dChisqOld >= psFS->dChisq ) {
         break;                       /* OK */
      } else {
                                      /* not, try again */
         psFS->dLambda *= 1E1; iLoop--;
                                      /* restore old parameters */
         fitFctBlkParBlkPut( psFFB, &sFPB, lFlag );
      }
   } while( iLoop>0 );
   psFS->dLambda *= 1E-1;

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }

   fitParBlkFree(&sFPB,lFlag);

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
